@@ -1,25 +1,35 @@
-/*  $Id$
-
-    Part of SWI-Prolog
+/*  Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (c)  1995-2016, University of Amsterdam
+    All rights reserved.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+    Redistribution and use in source and binary forms, with or without
+    modification, are permitted provided that the following conditions
+    are met:
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+    1. Redistributions of source code must retain the above copyright
+       notice, this list of conditions and the following disclaimer.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+    2. Redistributions in binary form must reproduce the above copyright
+       notice, this list of conditions and the following disclaimer in
+       the documentation and/or other materials provided with the
+       distribution.
+
+    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
+    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
+    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
+    ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+    POSSIBILITY OF SUCH DAMAGE.
 */
 
 #ifdef __WINDOWS__
@@ -319,7 +329,7 @@ setOSPrologFlags(void)
 
 
 char *
-findExecutable(const char *module, char *exe)
+findExecutable(const char *module, char *exe, size_t exelen)
 { int n;
   wchar_t wbuf[MAXPATHLEN];
   HMODULE hmod;
@@ -338,12 +348,9 @@ findExecutable(const char *module, char *exe)
 
   if ( (n = GetModuleFileNameW(hmod, wbuf, MAXPATHLEN)) > 0 )
   { wbuf[n] = EOS;
-    return _xos_long_file_name_toA(wbuf, exe, MAXPATHLEN);
+    return _xos_long_file_name_toA(wbuf, exe, exelen);
   } else if ( module )
-  { char buf[MAXPATHLEN];
-    PrologPath(module, buf, sizeof(buf));
-
-    strcpy(exe, buf);
+  { return PrologPath(module, exe, exelen);
   } else
     *exe = EOS;
 
@@ -597,7 +604,7 @@ pl_win_module_file(term_t module, term_t file)
 
   if ( !PL_get_chars(module, &m, CVT_ALL|CVT_EXCEPTION) )
     fail;
-  if ( (f = findExecutable(m, buf)) )
+  if ( (f = findExecutable(m, buf, sizeof(buf))) )
     return PL_unify_atom_chars(file, f);
 
   fail;
