@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2016, University of Amsterdam
+    Copyright (c)  1985-2017, University of Amsterdam
                               VU University Amsterdam
     All rights reserved.
 
@@ -50,9 +50,6 @@ static int  is_external(const char *rec, size_t len);
 #define RECORDA 0
 #define RECORDZ 1
 
-#define LOCK()   PL_LOCK(L_RECORD)
-#define UNLOCK() PL_UNLOCK(L_RECORD)
-
 #undef LD
 #define LD LOCAL_LD
 
@@ -105,6 +102,19 @@ lookupRecordList(word key)
 }
 
 
+static RecordRef
+firstRecordRecordList(RecordList rl)
+{ RecordRef record;
+
+  for(record = rl->firstRecord; record; record = record->next)
+  { if ( false(record->record, R_ERASED) )
+      return record;
+  }
+
+  return NULL;
+}
+
+
 static RecordList
 isCurrentRecordList(word key, int must_be_non_empty)
 { GET_LD
@@ -114,12 +124,10 @@ isCurrentRecordList(word key, int must_be_non_empty)
   { if ( must_be_non_empty )
     { RecordRef record;
 
-      LOCK();
-      for(record = rl->firstRecord; record; record = record->next)
-      { if ( false(record->record, R_ERASED) )
-	  break;
-      }
-      UNLOCK();
+      PL_LOCK(L_RECORD);
+      record = firstRecordRecordList(rl);
+      PL_UNLOCK(L_RECORD);
+
       return record ? rl : NULL;
     } else
     { return rl;
@@ -203,6 +211,11 @@ unallocRecordList(RecordList rl)
 
 #define dataRecord(r) ((char *)addPointer(r, SIZERECORD(r->flags)))
 
+typedef enum
+{ ENONE = 0,
+  EFAST_SERIALIZE
+} cerror;
+
 typedef struct
 { tmp_buffer code;			/* code buffer */
   tmp_buffer vars;			/* variable pointers */
@@ -210,6 +223,8 @@ typedef struct
   uint	     nvars;			/* # variables */
   int	     external;			/* Allow for external storage */
   int	     lock;			/* lock compiled atoms */
+  cerror     error;			/* generated error */
+  word	     econtext[1];		/* error context */
 } compile_info, *CompileInfo;
 
 #define	PL_TYPE_VARIABLE	(1)	/* variable */
@@ -402,12 +417,9 @@ addAtom(CompileInfo info, atom_t a)
 
       addAtomValue(info, ap);
     } else
-    { GET_LD
-      term_t t;
-
-      return ( (t=PL_new_term_ref()) &&
-		PL_put_atom(t, a) &&
-		PL_permission_error("fast_serialize", "blob", t) );
+    { info->error = EFAST_SERIALIZE;
+      info->econtext[0] = a;
+      return FALSE;
     }
   } else
   { addOpCode(info, PL_TYPE_ATOM);
@@ -710,6 +722,15 @@ compileTermToHeap__LD(term_t t,
   return record;
 }
 
+
+int
+variantRecords(const Record r1, const Record r2)
+{ return ( r1->size == r2->size &&
+	   memcpy(r1, r2, r1->size) == 0
+	 );
+}
+
+
 		 /*******************************
 		 *	 EXTERNAL RECORDS	*
 		 *******************************/
@@ -748,6 +769,22 @@ discard_record_data(record_data *data)
     discardBuffer(&data->hdr);
 }
 
+static int
+rec_error(CompileInfo info)
+{ switch(info->error)
+  { case EFAST_SERIALIZE:
+    { GET_LD
+      term_t t;
+
+      return ( (t=PL_new_term_ref()) &&
+		PL_put_atom(t, info->econtext[0]) &&
+		PL_permission_error("fast_serialize", "blob", t) );
+    }
+    default:
+      assert(0);
+  }
+}
+
 
 static int
 compile_external_record(term_t t, record_data *data ARG_LD)
@@ -801,12 +838,12 @@ general:
   initTermAgenda(&agenda, 1, p);
   rc = compile_term_to_heap(&agenda, &data->info PASS_LD);
   clearTermAgenda(&agenda);
-  if ( !rc )
-    return FALSE;
   if ( data->info.nvars == 0 )
     first |= REC_GROUND;
   restoreVars(&data->info);
   unvisit(PASS_LD1);
+  if ( !rc )
+    return rec_error(&data->info);
   scode = (int)sizeOfBuffer(&data->info.code);
 
   initBuffer(&data->hdr);
@@ -913,7 +950,8 @@ PRED_IMPL("fast_term_serialized", 2, fast_term_serialized, 0)
     } else
     { return FALSE;
     }
-  } else if ( PL_get_string_chars(string, &rec, &len) )
+  } else if ( PL_get_nchars(string, &len, &rec,
+			    CVT_STRING|BUF_RING|REP_ISO_LATIN_1|CVT_EXCEPTION) )
   { term_t tmp;
 
     return ( (tmp = PL_new_term_ref()) &&
@@ -921,7 +959,7 @@ PRED_IMPL("fast_term_serialized", 2, fast_term_serialized, 0)
 	     PL_recorded_external(rec, tmp) &&
 	     PL_unify(term, tmp) );
   } else
-  { return PL_type_error("string", string);
+  { return FALSE;
   }
 }
 
@@ -1868,9 +1906,12 @@ real atom.
 bool
 unifyKey(term_t key, word val)
 { if ( isAtom(val) || isTaggedInt(val) )
-    return _PL_unify_atomic(key, val);
+  { return _PL_unify_atomic(key, val);
+  } else
+  { GET_LD
 
-  return PL_unify_functor(key, (functor_t) val);
+    return PL_unify_functor(key, (functor_t) val);
+  }
 }
 
 
@@ -1970,7 +2011,7 @@ record(term_t key, term_t term, term_t ref, int az)
     return FALSE;
   }
 
-  LOCK();
+  PL_LOCK(L_RECORD);
   l = lookupRecordList(k);
   r->list = l;
 
@@ -1989,7 +2030,7 @@ record(term_t key, term_t term, term_t ref, int az)
     l->lastRecord = r;
   }
 
-  UNLOCK();
+  PL_UNLOCK(L_RECORD);
 
   succeed;
 }
@@ -2007,6 +2048,16 @@ PRED_IMPL("recordz", va, recordz, 0)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+recorded/2,3. The state enumerates keys using the  `e` member if the key
+is unbound on  entry.  The  `r`  member   is  the  current  record.
+
+Enumeration first scans the records and then, if `e` is set, advanced to
+the next key.
+
+All manipulation on the state is done whild holding L_RECORD.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 typedef struct
 { TableEnum e;				/* enumerating over keys */
   RecordRef r;				/* current record */
@@ -2025,15 +2076,46 @@ save_state(recorded_state *state)
   }
 }
 
+/* MT: must hold L_RECORD */
+
 static void
 free_state(recorded_state *state)
 { if ( state->e )
     freeTableEnum(state->e);
+  if ( state->r )
+  { RecordList rl = state->r->list;
+
+    if ( --rl->references == 0 && true(rl, RL_DIRTY) )
+      cleanRecordList(rl);
+  }
   if ( state->saved )
     freeForeignState(state, sizeof(*state));
 }
 
 
+/* set state to the next non-erased record.  Cleanup the record
+   list if we reached the end.
+*/
+
+static RecordRef
+advance_state(recorded_state *state)
+{ RecordRef r = state->r;
+
+  do
+  { if ( !r->next )
+    { RecordList rl = r->list;
+
+      if ( --rl->references == 0 && true(rl, RL_DIRTY) )
+	cleanRecordList(rl);
+    }
+    r = r->next;
+  } while ( r && true(r->record, R_ERASED) );
+
+  state->r = r;
+  return r;
+}
+
+
 static
 PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 { PRED_LD
@@ -2053,7 +2135,7 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
       { RecordRef record;
 
 	if ( PL_get_recref(ref, &record) )
-	{ LOCK();
+	{ PL_LOCK(L_RECORD);
 	  if ( unifyKey(key, record->list->key) )
 	  { term_t copy = PL_new_term_ref();
 
@@ -2064,7 +2146,7 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 	      rc = PL_unify(term, copy);
 	  } else
 	    rc = FALSE;
-	  UNLOCK();
+	  PL_UNLOCK(L_RECORD);
 
 	  return rc;
 	}
@@ -2074,56 +2156,50 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
       memset(state, 0, sizeof(*state));
       if ( PL_is_variable(key) )
       { state->e = newTableEnum(GD->recorded_db.record_lists);
+	PL_LOCK(L_RECORD);
       } else if ( getKeyEx(key, &k PASS_LD) )
       { RecordList rl;
 
 	if ( !(rl = isCurrentRecordList(k, TRUE)) )
 	  return FALSE;
+	PL_LOCK(L_RECORD);
 	rl->references++;
 	state->r = rl->firstRecord;
       } else
       { return FALSE;
       }
-      LOCK();
       break;
     }
     case FRG_REDO:
     { state = CTX_PTR;
-      LOCK();
+      PL_LOCK(L_RECORD);
       break;
     }
     case FRG_CUTTED:
     { state = CTX_PTR;
-
-      if ( state->r )
-      { RecordList rl = state->r->list;
-
-	LOCK();
-	if ( --rl->references == 0 && true(rl, RL_DIRTY) )
-	  cleanRecordList(rl);
-	UNLOCK();
-      }
+      PL_LOCK(L_RECORD);
       free_state(state);
+      PL_UNLOCK(L_RECORD);
     }
-      /* FALLTHROUGH */
+    /*FALLTHROUGH*/
     default:
       succeed;
   }
 
+  /* Now holding L_RECORD */
   if ( (fid = PL_open_foreign_frame()) )
   { int answered = FALSE;
+    term_t copy = 0;
 
     while( !answered )
-    { for( ; state->r; state->r = state->r->next )
-      { term_t copy;
-	RecordRef record;
+    { for( ; state->r; advance_state(state) )
+      { RecordRef record;
 
       next:
 	record = state->r;
-	if ( true(record->record, R_ERASED) )
-	  continue;
 
-	copy = PL_new_term_ref();
+	if ( !copy && !(copy = PL_new_term_ref()) )
+	  goto error;
 	if ( (rc=copyRecordToGlobal(copy, record->record, ALLOW_GC PASS_LD)) < 0 )
 	{ raiseStackOverflow(rc);
 	  goto error;
@@ -2144,14 +2220,9 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 
 	if ( record->next )
 	{ state->r = record->next;
-	  UNLOCK();
+	  PL_UNLOCK(L_RECORD);
 	  PL_close_foreign_frame(fid);
 	  ForeignRedoPtr(save_state(state));
-	} else
-	{ RecordList rl = record->list;
-
-	  if ( --rl->references == 0 && true(rl, RL_DIRTY) )
-	    cleanRecordList(rl);
 	}
       }
 
@@ -2160,10 +2231,11 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
 
 	while(advanceTableEnum(state->e, &sk, &sv))
 	{ RecordList rl = sv;
+	  RecordRef r;
 
-	  if ( rl->firstRecord )
+	  if ( (r=firstRecordRecordList(rl)) )
 	  { rl->references++;
-	    state->r = rl->firstRecord;
+	    state->r = r;
 	    if ( answered )
 	      break;
 	    goto next;			/* try next list */
@@ -2172,12 +2244,15 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
       }
 
       if ( answered )
-      { UNLOCK();
-	PL_close_foreign_frame(fid);
+      { PL_close_foreign_frame(fid);
 	if ( state->e )
+	{ PL_UNLOCK(L_RECORD);
 	  ForeignRedoPtr(save_state(state));
-	else
+	} else
+	{ free_state(state);
+	  PL_UNLOCK(L_RECORD);
 	  return TRUE;
+	}
       }
 
       break;
@@ -2187,8 +2262,9 @@ PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
     PL_close_foreign_frame(fid);
   }
 
-  UNLOCK();
   free_state(state);
+  PL_UNLOCK(L_RECORD);
+
   return FALSE;
 }
 
@@ -2262,10 +2338,11 @@ PRED_IMPL("erase", 1, erase, 0)
     return retractClauseDefinition(def, clause);
   } else
   { RecordRef r = ptr;
+    int rc;
 
-    callEventHook(PLEV_ERASED_RECORD, r);
+    rc = callEventHook(PLEV_ERASED_RECORD, r);
 
-    LOCK();
+    PL_LOCK(L_RECORD);
     l = r->list;
     if ( l->references )		/* a recorded has choicepoints */
     { set(r->record, R_ERASED);
@@ -2273,8 +2350,8 @@ PRED_IMPL("erase", 1, erase, 0)
     } else
     { remove_record(r);
     }
-    UNLOCK();
-    return TRUE;
+    PL_UNLOCK(L_RECORD);
+    return rc;
   }
 }
 
